import pandas as pd
import numpy as np
import os
from collections import Counter
from scipy.cluster.hierarchy import linkage, fcluster
from scipy.spatial.distance import squareform


def find_correlation_clusters(corr,corr_thresh):
    dissimilarity = 1.0 - corr
    hierarchy = linkage(squareform(dissimilarity), method='single')
    diss_thresh = 1.0 - corr_thresh
    labels = fcluster(hierarchy, diss_thresh, criterion='distance')
    return labels

def relabel_clusters(labels,metric_columns):
    cluster_count = Counter(labels)
    cluster_order = {cluster[0]: idx for idx, cluster in enumerate(cluster_count.most_common())}
    relabeled_clusters = [cluster_order[l] for l in labels]
    relabled_count = Counter(relabeled_clusters)
    labeled_column_df = pd.DataFrame({'group': relabeled_clusters, 'column': metric_columns}).sort_values(
        ['group', 'column'], ascending=[True, True])
    return labeled_column_df, relabled_count

def make_load_matrix(labeled_column_df,metric_columns,relabled_count,corr):
    load_mat = np.zeros((len(metric_columns), len(relabled_count)))
    for row in labeled_column_df.iterrows():
        orig_col = metric_columns.index(row[1][1])
        if relabled_count[row[1][0]]>1:
            load_mat[orig_col, row[1][0]] = 1.0/  (np.sqrt(corr) * float(relabled_count[row[1][0]])  )
        else:
            load_mat[orig_col, row[1][0]] = 1.0

    is_group = load_mat.astype(bool).sum(axis=0) > 1
    column_names=['metric_group_{}'.format(d + 1) if is_group[d]
                      else labeled_column_df.loc[labeled_column_df['group']==d,'column'].iloc[0]
                      for d in range(0, load_mat.shape[1])]
    loadmat_df = pd.DataFrame(load_mat, index=metric_columns, columns=column_names)
    loadmat_df['name'] = loadmat_df.index
    sort_cols = list(loadmat_df.columns.values)
    sort_order = [False] * loadmat_df.shape[1]
    sort_order[-1] = True
    loadmat_df = loadmat_df.sort_values(sort_cols, ascending=sort_order)
    loadmat_df = loadmat_df.drop('name', axis=1)
    return loadmat_df

def save_load_matrix(data_set_path,loadmat_df, labeled_column_df):
    save_path = data_set_path.replace('.csv', '_load_mat.csv')
    print('saving loadings to ' + save_path)
    loadmat_df.to_csv(save_path)
    save_path = data_set_path.replace('.csv', '_groupmets.csv')
    print('saving metric groups to ' + save_path)
    group_lists=['|'.join(labeled_column_df[labeled_column_df['group']==g]['column'])
                    for g in set(labeled_column_df['group'])]
    pd.DataFrame(group_lists,index=loadmat_df.columns.values,columns=['metrics']).to_csv(save_path)

def find_metric_groups(data_set_path,group_corr_thresh=0.5):

    score_save_path=data_set_path.replace('.csv','_scores.csv')
    assert os.path.isfile(score_save_path),'You must run listing 5.3 or 7.5 to save metric scores first'
    score_data = pd.read_csv(score_save_path,index_col=[0,1])
    score_data.drop('is_churn',axis=1,inplace=True)
    metric_columns = list(score_data.columns.values)

    labels = find_correlation_clusters(score_data.corr(),group_corr_thresh)
    labeled_column_df, relabled_count = relabel_clusters(labels,metric_columns)
    loadmat_df = make_load_matrix(labeled_column_df, metric_columns, relabled_count,group_corr_thresh)
    save_load_matrix(data_set_path,loadmat_df,labeled_column_df)



import json
from collections import OrderedDict
import re

conf_path = '../../listings/conf/socialnet7_listings.json'

with open(conf_path, 'r') as myfile:
    param_dict = json.loads(myfile.read())

new_conf = OrderedDict()

for chap in param_dict.keys():
    new_conf[chap] = OrderedDict()
    for listing in param_dict[chap].keys():
        if listing=='params':
            new_conf[chap]['defaults']=param_dict[chap]['params']
            continue
        # print(listing)
        parts=listing.split('_')
        print(parts)
        list_type=parts[0]
        chap_num=parts[1]
        list_num=parts[2]
        if parts[3].isdigit():
            vers_num=parts[3]
            list_name='_'.join(parts[4:])
        else:
            vers_num = None
            list_name='_'.join(parts[3:])

        list_num_name = f'list{list_num}'
        if not list_num_name in new_conf[chap]:
            new_conf[chap][list_num_name] = OrderedDict()
            new_conf[chap][list_num_name]['name']=list_name

        if list_type == 'insert':
            if 'insert' not in new_conf[chap][list_num_name]:
                new_conf[chap][list_num_name]['insert']=OrderedDict()
            param_target = new_conf[chap][list_num_name]['insert']
        else:
            param_target = new_conf[chap][list_num_name]

        if vers_num:
            vers_num_name = f'v{vers_num}'
            param_target[vers_num_name] = param_dict[chap][listing]
            # if there are defaults, remove version params matching defaults
            if 'params' in param_target:
                for k in list(param_target[vers_num_name].keys()):
                    if k in param_target['params']:
                        if param_target[vers_num_name][k]==param_target['params'][k]:
                            param_target[vers_num_name].pop(k)
        else:
            param_target['params'] =param_dict[chap][listing]


save_path = '../../listings/conf/socialnet7_listings_v2.json'

with open(save_path, 'w') as myfile:
    json.dump(new_conf,myfile,indent=4)



import pandas as pd

def dataset_correlation_matrix(data_set_path):

    churn_data = pd.read_csv(data_set_path,index_col=[0,1])
    churn_data = churn_data.reindex(sorted(churn_data.columns), axis=1)
    corr_df = churn_data.corr()

    save_name = data_set_path.replace('.csv', '_correlation_matrix.csv')
    corr_df.to_csv(save_name)
    print('Saved correlation matrix to ' + save_name)



from listing_5_2_dataset_stats import dataset_stats
from listing_8_6_clipped_scores import clipped_scores
from listing_6_4_find_metric_groups import find_metric_groups
from listing_6_3_apply_metric_groups import apply_metric_groups

def prepare_data(data_set_path):
    dataset_stats(data_set_path)
    clipped_scores(data_set_path)
    find_metric_groups(data_set_path)
    apply_metric_groups(data_set_path)



import pandas as pd
import ntpath
import numpy as np

from sklearn.model_selection import GridSearchCV, TimeSeriesSplit
from sklearn.metrics import make_scorer
from sklearn.linear_model import LogisticRegression
import matplotlib.pyplot as plt

from listing_8_2_logistic_regression import prepare_data
from listing_9_2_top_decile_lift import calc_lift


def crossvalidate(data_set_path,n_test_split):

    X,y = prepare_data(data_set_path,as_retention=False)
    tscv = TimeSeriesSplit(n_splits=n_test_split)
    score_models = {'lift': make_scorer(calc_lift, needs_proba=True), 'AUC': 'roc_auc'}
    retain_reg = LogisticRegression(penalty='l1', solver='liblinear', fit_intercept=True)
    test_params = {'C' : [0.64, 0.32, 0.16, 0.08, 0.04, 0.02, 0.01, 0.005, 0.0025]}
    gsearch = GridSearchCV(estimator=retain_reg,scoring=score_models, cv=tscv, verbose=1,
                           return_train_score=False,  param_grid=test_params, refit=False)
    gsearch.fit(X,y)

    result_df = pd.DataFrame(gsearch.cv_results_)
    result_df['n_weights']= test_n_weights(X,y,test_params)
    result_df.to_csv(data_set_path.replace('.csv', '_crossval.csv'), index=False)
    plot_regression_test(data_set_path,result_df)

def test_n_weights(X,y,test_params):
    n_weights=[]
    for c in test_params['C']:
        lr = LogisticRegression(penalty='l1',C=c, solver='liblinear', fit_intercept=True)
        res=lr.fit(X,~y)
        n_weights.append(res.coef_[0].astype(bool).sum(axis=0))
    return n_weights

def plot_regression_test(data_set_path, result_df):
    result_df['plot_C']=result_df['param_C'].astype(str)
    plt.figure(figsize=(4,6))
    plt.rcParams.update({'font.size':8})
    one_subplot(result_df,1,'mean_test_AUC',ylim=(0.5,0.8),ytick=.05)
    plt.title(ntpath.basename(data_set_path).replace('_dataset.csv',' Cross Validation'))
    one_subplot(result_df,2,'mean_test_lift',ylim=(1, 6),ytick=0.5)
    one_subplot(result_df,3,'n_weights', ylim=(0, int(1 + result_df['n_weights'].max())),ytick=2)
    plt.xlabel('Regression C Param')
    plt.savefig(data_set_path.replace('.csv', '_crossval_regression.png'))
    plt.close()

def one_subplot(result_df,plot_n,var_name,ylim,ytick):
    ax = plt.subplot(3,1,plot_n)
    ax.plot('plot_C', var_name, data=result_df, marker='o', label=var_name)
    plt.ylim(ylim[0],ylim[1])
    plt.yticks(np.arange(ylim[0], ylim[1], step=ytick))
    plt.legend()
    plt.grid()



from sklearn.linear_model import LogisticRegression
from listing_8_2_logistic_regression import prepare_data, save_regression_model
from listing_8_2_logistic_regression import save_regression_summary, save_dataset_predictions

def regression_cparam(data_set_path, C_param):
    X,y = prepare_data(data_set_path)
    retain_reg = LogisticRegression( C=C_param, penalty='l1', solver='liblinear', fit_intercept=True)
    retain_reg.fit(X, y)
    c_ext = '_c{:.3f}'.format(C_param)
    save_regression_summary(data_set_path,retain_reg,ext=c_ext)
    save_regression_model(data_set_path,retain_reg,ext=c_ext)
    save_dataset_predictions(data_set_path,retain_reg,X,ext=c_ext)




import pandas as pd
import os
from listing_7_5_fat_tail_scores import transform_fattail_columns, transform_skew_columns

def clip_hi_cols(data, hi_vals):
    for col in hi_vals.index.values:
        data.loc[data[col] > hi_vals[col],col] = hi_vals[col]

def clip_lo_cols(data, lo_vals):
    for col in lo_vals.index.values:
        data.loc[data[col] < lo_vals[col],col] = lo_vals[col]

def clipped_scores(data_set_path,skew_thresh=4.0):

    churn_data = pd.read_csv(data_set_path)
    churn_data.set_index(['account_id','observation_date'],inplace=True)
    data_scores = churn_data.copy()
    data_scores.drop('is_churn',axis=1,inplace=True)

    stat_path = data_set_path.replace('.csv', '_summarystats.csv')
    assert os.path.isfile(stat_path),'You must running listing 5.2 first to generate stats'
    stats = pd.read_csv(stat_path,index_col=0)
    stats.drop('is_churn',inplace=True)

    clip_hi_cols(data_scores, stats['99pct'])
    clip_lo_cols(data_scores, stats['1pct'])

    skewed_columns=(stats['skew']>skew_thresh) & (stats['min'] >= 0)
    transform_skew_columns(data_scores,skewed_columns[skewed_columns].keys())

    fattail_columns=(stats['skew']>skew_thresh) & (stats['min'] < 0)
    transform_fattail_columns(data_scores,fattail_columns[fattail_columns].keys())

    mean_vals = data_scores.mean()
    std_vals = data_scores.std()
    data_scores=(data_scores-mean_vals)/std_vals

    data_scores['is_churn']=churn_data['is_churn']

    score_save_path=data_set_path.replace('.csv','_scores.csv')
    data_scores.to_csv(score_save_path,header=True)
    print('Saving results to %s' % score_save_path)

    param_df = pd.DataFrame({'skew_score': skewed_columns,
                                 'fattail_score': fattail_columns,
                                 'mean': mean_vals,
                                 'std': std_vals})
    param_save_path=data_set_path.replace('.csv','_score_params.csv')
    param_df.to_csv(param_save_path,header=True)
    print('Saving params to %s' % param_save_path)



import os
import pickle
import shap
import matplotlib.pyplot as plt
from listing_8_4_rescore_metrics import reload_churn_data

def shap_explain_logreg(data_set_path):

    pickle_path = data_set_path.replace('.csv', '_logreg_model_churn.pkl')
    assert os.path.isfile(pickle_path), 'You must run listing 8.2 with param as_retention=False to save a churn logistic regression model first'
    with open(pickle_path, 'rb') as fid:
        logreg_model = pickle.load(fid)

    current_df = reload_churn_data(data_set_path,'current_groupscore','8.4',is_customer_data=True)
    explainer_mask = shap.maskers.Impute(data=current_df)
    explainer = shap.LinearExplainer(logreg_model, masker=explainer_mask, model_output='probability')
    shap_values = explainer(current_df)
    shap.summary_plot(shap_values, current_df, show=False)
    save_file = data_set_path.replace('.csv', '_shap_summary_logreg.png')
    print(f'Saving SHAP Explanation to {save_file}')
    plt.tight_layout()
    plt.savefig(save_file, format='png')
    plt.close()



import setuptools
import pkg_resources
import pathlib

# https://stackoverflow.com/questions/49689880/proper-way-to-parse-requirements-file-after-pip-upgrade-to-pip-10-x-x
with pathlib.Path('requirements.txt').open() as requirements_txt:
    install_requires = [
        str(requirement)
        for requirement
        in pkg_resources.parse_requirements(requirements_txt)
    ]

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="fightchurn",
    version="0.1.1",
    author="Carl Gold",
    author_email="carl24k@fightchurnwithdata.com",
    description="Code from the book Fighting Churn With Data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/carl24k/fight-churn",
    project_urls={
        "Bug Tracker": "https://github.com/carl24k/fight-churn/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_data={'fightchurn' : ['fightchurn/listings/conf/*.json',
                                  'fightchurn/listings/*/*.sql']},
    include_package_data=True,
    packages=['fightchurn',
              'fightchurn.listings',
              'fightchurn.listings.conf',
              'fightchurn.listings.chap3',
              'fightchurn.listings.chap5',
              'fightchurn.listings.chap6',
              'fightchurn.listings.chap7',
              'fightchurn.listings.chap8',
              'fightchurn.listings.chap9',
              'fightchurn.listings.chap10'],
    scripts=['fightchurn/run_churn_listing.py'],
    python_requires=">=3.9",
    install_requires= install_requires
)



import os
import pandas as pd
from listing_6_4_find_metric_groups import find_correlation_clusters, relabel_clusters, make_load_matrix, save_load_matrix

def find_groups_skip_cats(data_set_path,group_corr_thresh=0.5):

    score_save_path=data_set_path.replace('.csv','_scores.csv')
    assert os.path.isfile(score_save_path),'You must run listing 5.3 or 7.5 to save metric scores first'
    score_data = pd.read_csv(score_save_path,index_col=[0,1])
    score_data.drop('is_churn',axis=1,inplace=True)

    binary_cols = [col for col in score_data if len(score_data[col].unique())==2]
    score_data.drop(binary_cols,axis=1,inplace=True)

    metric_columns = list(score_data.columns.values)

    labels = find_correlation_clusters(score_data.corr(),group_corr_thresh)
    labeled_column_df, relabled_count = relabel_clusters(labels,metric_columns)
    loadmat_df = make_load_matrix(labeled_column_df, metric_columns, relabled_count,group_corr_thresh)

    n_non_bin = loadmat_df.shape[1]
    n_bin = len(binary_cols)
    new_label_rows = pd.DataFrame(data=list(zip(range(n_non_bin,n_non_bin+n_bin),binary_cols)),
                                  columns=labeled_column_df.columns)
    labeled_column_df = labeled_column_df.append(new_label_rows)

    for c in binary_cols:
        loadmat_df[c]=0.0
    new_mat_rows = pd.DataFrame(data=0.0,index=binary_cols,columns=loadmat_df.columns)
    loadmat_df = loadmat_df.append(new_mat_rows)
    for c in binary_cols:
        loadmat_df.at[c,c]=1.0

    save_load_matrix(data_set_path,loadmat_df,labeled_column_df)
